import configparser
import json
from collections.abc import Iterable, MutableMapping
from configparser import ConfigParser
from copy import deepcopy
from json import JSONDecodeError, JSONEncoder
from logging import Logger
from pathlib import Path

from lynceus.core.config import CONFIG_JSON_DUMP_KEY_END_KEYWORD
from lynceus.lynceus_exceptions import LynceusConfigError
from lynceus.utils import lookup_root_path


# pylint: disable=too-many-public-methods
class LynceusConfig(MutableMapping):
    """A dictionary that applies an arbitrary key-altering
       function before accessing the keys.

       IMPORTANT: when a key does not exist, it is automatically created to ease definition like ['X']['Y']['Z'] = 'XXX'
       BUT: it means the 'in' operator should NOT be used, because it will always return True, because the checked key will be created if not existing,
        for such situation, you should use the has_section() method.
       """

    UNDEFINED_VALUE: str = '$*_UNDEFINED_*$'

    def __init__(self, *args, **kwargs):
        self.__store = {}
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        # Creates automatically sub dictionary if needed.
        if LynceusConfig._keytransform(key) not in self.__store:
            self.__store[LynceusConfig._keytransform(key)] = {}

        return self.__store[LynceusConfig._keytransform(key)]

    def __setitem__(self, key, value):
        self.__store[LynceusConfig._keytransform(key)] = LynceusConfig._valuetransform(value)

    def __delitem__(self, key):
        del self.__store[LynceusConfig._keytransform(key)]

    def __iter__(self):
        return iter(self.__store)

    def __len__(self):
        return len(self.__store)

    @staticmethod
    def _keytransform(key):
        return key

    @staticmethod
    def _valuetransform(value):
        if isinstance(value, LynceusConfig):
            # pylint: disable=protected-access
            return value.__store

        return value

    def copy(self):
        lynceus_config_copy: LynceusConfig = LynceusConfig()
        # pylint: disable=protected-access,unused-private-member
        lynceus_config_copy.__store = deepcopy(self.__store)
        return lynceus_config_copy

    def as_dict(self):
        return deepcopy(self.__store)

    @staticmethod
    def dump_config_parser(config_parser: ConfigParser):
        result: str = ''
        for config_section in config_parser.sections():
            result += f'[{config_section}]\n'
            for key, value in config_parser.items(config_section):
                result += f'\t{key}={value}\n'
        return result

    def get_config(self, section: str, key: str, *, default: object = UNDEFINED_VALUE):
        try:
            return self[section][key]
        except KeyError as error:
            # Safe-guard: raise an error if configuration not found and no default specified.
            if default == self.UNDEFINED_VALUE:
                raise LynceusConfigError(f'Configuration [\'{section}\'][\'{key}\'] does not exist in your configuration.') from error
            return default

    def check_config_exists(self, section: str, key: str, *, transform_func=None):
        """
        Returns True if the configuration option with specified Key under specified section exists, False otherwise.
        If the value exists, system executes the optional func if specified, on the value (can be a cast, or any process function);
        caller need to manage potential exception during this execution.

        :param section:
        :param key:
        :param transform_func: (optional) **value** transformation function (can be a cast, or any process function)
        :return:
        """
        value = self.get_config(section, key, default=None)
        if value is None:
            return False

        # Checks if there is a trasnformation to apply.
        if transform_func:
            # Transforms (and registers) the value for further usage.
            self[section][key] = transform_func(value)

        # Returns the config option exists.
        return True

    @staticmethod
    def to_bool(value: str | None) -> bool | None:
        if value is None:
            return None
        return value in ['1', 'True', True]

    def is_bool_config_enabled(self, section: str, key: str, *, default: bool = False):
        return LynceusConfig.to_bool(self.get_config(section=section, key=key, default=default))

    def get_config_as_dict(self, section: str, key: str):
        # TODO: enhance it to work from API version which will be a string ...
        return self.get_config(section=section, key=key)

    def has_section(self, section: str):
        # Important: we must check the specified section on a list of keys, and not on the result of keys(), otherwise tha automatic
        #  key addition will be performed in __getitem__() method, and this method will always answer True.
        return section in list(self.keys())

    @staticmethod
    def __cleanse_value(value: str):
        return value if not isinstance(value, str) else str(value).strip("'").strip('"')

    @staticmethod
    # pylint: disable=too-many-branches
    def recursive_merge(dict1: dict | MutableMapping, dict2: dict | MutableMapping, *,
                        override_section: bool = False, override_value: bool = False, key_transform_func=None):
        """
        Merges all data from specified dict2, to the specified dict1, without removing/overriding any existing value, when possible.

        :param dict1: (aka destination dictionary)
        :param dict2: (aka source dictionary)
        :param override_section: overrides existing section (section=dict/mapping), or merges all existing key, value pairs of existing section (default)
        :param override_value: overrides existing value (value=all type which is not a section) or raises an exception (default)
        :param key_transform_func: optional **key** transformation function (e.g. useful to merge project template LynceusMetadata configuration in the project to score with
                    updated key, to state they belong to the template and not the project itself).

        For instance (very simplified):
        [General]                           Top level **section**
        logger.console.enable=True              simple key=value

        
        

        [MyKey2]                            Top level **section**
            [MyInnerKey]                        Second level **section**
                MyLeafKey=MyValue                   simple key=value
        """
        for key, value in dict2.items():
            # Checks if the corresponding key already exists in destination dict.
            # Important: do NOT use the in operator if first argument is a LynceusConfig, to avoid automatic creation of section here.
            if (isinstance(dict1, LynceusConfig) and not dict1.has_section(key)) or key not in dict1:
                dict1[key] = value
                continue

            # Checks if existing destination is iterable but NOT a map/mapping.
            if isinstance(dict1[key], Iterable) and not isinstance(dict1[key], (dict, MutableMapping, str)):
                # Checks if "new" value is iterable
                # N.B.: If it is NOT the case, it will be managed as a no map-like one with further instructions.
                if isinstance(value, Iterable) and not isinstance(value, str):
                    # Adapts merge according to concrete type.
                    if isinstance(dict1[key], list):
                        dict1[key].extend(value)
                    elif isinstance(dict1[key], set):
                        dict1[key] |= set(value)
                    else:
                        raise NotImplementedError(f'Recursive merge is not implemented for config option of type {type(dict1[key])=}.')
                    continue

            # Checks if existing destination is a map-like.
            if not isinstance(dict1[key], (dict, MutableMapping)):
                # Transforms the leaf key if requested.
                if key_transform_func:
                    key = key_transform_func(key)

                # Checks if override_value is allowed.
                if override_value:
                    dict1[key] = value
                    continue

                # Checks if it is not already the same value.
                if dict1[key] != value:
                    raise LynceusConfigError(f'Merge of "({key}, {value})" pair would override existing "({key}, {dict1[key]})" pair, whose value is NOT a dictionary/mapping.')

                continue

            # Overrides section if requested.
            if override_section:
                dict1[key] = value
                continue

            # Checks if source value is a map/mapping, otherwise.
            if isinstance(value, (dict, MutableMapping)):
                # Calls recursively this method on them.
                LynceusConfig.recursive_merge(dict1[key], value, override_value=override_value, override_section=override_section, key_transform_func=key_transform_func)
                continue

            # Updates the destination with the current pair.
            dict1[key].update({key: value})

    def merge(self, config_map: dict[str, any], *,
              set_only_as_default_if_not_exist: bool = False, override_section: bool = False, override_value: bool = False, key_transform_func=None):
        """
        Merges specified dict into this LynceusConfig instance.
        If set_only_as_default_if_not_exist is False (default), all existing inner dict-like values will be merged with new
         (key; value) pair if already exists, otherwise, the new values will only be considred as
         default value, and will only be set, if corresponding key does not already exists.

        :param config_map: the config map to merge into this instance
        :param set_only_as_default_if_not_exist: False (default) if existing value must be merged, True to
         only set value(s) if the corresponding key(s) do(es) not already exist (thus NOT overriding existing (key,value) pairs).
        :param override_section: overrides existing section (section=dict/mapping), or merges all existing key, value pairs of existing section (default)
        :param override_value: overrides existing value (value=all type which is not a section) or raises an exception (default)
        :param key_transform_func: optional **key** transformation function (e.g. useful to merge project template LynceusMetadata configuration in the project to score with
                    updated key, to state they belong to the template and not the project itself).
        """

        if not isinstance(self, LynceusConfig):
            raise LynceusConfigError(f'Self should be LynceusConfig, but is {type(self)}')

        # Processes automatically specified config_map to manage optional json dump options.
        config_map = LynceusConfig.load_json_params_from_lynceus_config(config_map).as_dict()

        if set_only_as_default_if_not_exist:
            if key_transform_func:
                raise NotImplementedError('key_transform_func option can not be used with set_only_as_default_if_not_exist option while merging config map.')

            # Uses the Python 3.9 ability to update Dictionary (considering existing values as the
            #  precedence ones).
            # TODO: it may be needed to iterate on each (key, value) pairs and recursively execute the | operator on them if value is dict-like instance ...
            self.__store = config_map | self.__store
            return

        # In this version, we need to merge everything, keeping existing inner (key; value) pairs.
        LynceusConfig.recursive_merge(self, config_map, override_section=override_section, override_value=override_value, key_transform_func=key_transform_func)

    def merge_from_lynceus_config(self, lynceus_config, *, override_section: bool = False, override_value: bool = False):
        # pylint: disable=protected-access
        self.merge(lynceus_config.__store, override_section=override_section, override_value=override_value)

    def merge_from_config_parser(self, config_parser: ConfigParser, *, override_section: bool = False, key_transform_func=None):
        """
        Merges configuration coming from specified configParser.

        :param config_parser:
        :param override_section: if True, resets all inner (key; value) pairs of existing section (e.g. top level key) in both this instance, and specified configParse,
                    merge everything otherwise (Default: False)
        :param key_transform_func: optional **key** transformation function (e.g. useful to merge project template LynceusMetadata configuration in the project to score with
                    updated key, to state they belong to the template and not the project itself).
        """
        # Merges each config parser section, as a dictionary, in this instance of Lynceus config.
        config_parser_to_dict = {config_section: {key: self.__cleanse_value(value) for key, value in config_parser.items(config_section)}
                                 for config_section in config_parser.sections()}
        self.merge(config_parser_to_dict, override_section=override_section, override_value=True, key_transform_func=key_transform_func)

    def lookup_configuration_file_and_update_from_it(self, config_file_name: str, *,
                                                     must_exist: bool = True,
                                                     root_path: Path = Path().resolve(),
                                                     logger: Logger | None = None,
                                                     **kwargs):
        try:
            # Retrieves the complete path of the configuration file, if exists.
            config_file_path = lookup_root_path(config_file_name, root_path=root_path) / Path(config_file_name)

            # Merges it in this instance.
            self.update_from_configuration_file(config_file_path, **kwargs)

            if logger:
                logger.info(f'Successfully load and merge configuration options from file "{config_file_name}".')
        except FileNotFoundError:
            # Raises FileNotFoundError only if configuration file should exist.
            if must_exist:
                raise

            if logger:
                logger.info(f'Configuration file "{config_file_name}" does not exist, so it will not be loaded (nor an error because {must_exist=}).')

    def update_from_configuration_file(self, file_path: Path, *,
                                       override_section: bool = False, key_transform_func=None):
        """
        Merges configuration coming from configuration file whose path is specified.

        :param file_path: path of the configuration file.
        :param override_section: if True, resets all inner (key; value) pairs of existing section (e.g. top level key) in both this instance, and specified configParse,
                    merge everything otherwise (Default: False).
        :param key_transform_func: optional **key** transformation function (e.g. useful to merge project template LynceusMetadata configuration in the project to score with
                    updated key, to state they belong to the template and not the project itself).
        """
        if not file_path.exists():
            raise FileNotFoundError(f'Specified "{file_path}" configuration file path does not exist.')

        # Reads configuration file.
        config_parser: configparser.ConfigParser = configparser.RawConfigParser()
        config_parser.read(str(file_path), encoding='utf8')

        # Merges it into this instance.
        self.merge_from_config_parser(config_parser, override_section=override_section, key_transform_func=key_transform_func)

    @staticmethod
    def format_config(config) -> dict[str, str]:
        def obfuscator(key, value):
            return '<secret>' if ('secret' in key or 'password' in key or 'pwd' in key or 'token' in key) else value

        return {key: obfuscator(key, value) if not isinstance(value, dict) else LynceusConfig.format_config(value)
                for key, value in dict(config).items()}

    @staticmethod
    def format_dict_to_string(dict_to_convert: MutableMapping, indentation_level: int = 0) -> str:
        return '\n'.join('\t' * indentation_level + f'{key}={str(value)}'
                         if not isinstance(value, dict) and not isinstance(value, LynceusConfig)
                         else '\t' * indentation_level + f'[{key}]\n' + LynceusConfig.format_dict_to_string(value, indentation_level + 1)
                         for key, value in dict_to_convert.items())

    def dump_for_config_file(self) -> str:
        return LynceusConfig.dump_to_config_str(self.__store)

    @staticmethod
    def dump_to_config_str(value) -> str:
        return json.dumps(value, cls=LynceusConfigJSONEncoder)

    @staticmethod
    def load_from_config_str(lynceus_config_dump: str | dict[str, any]):
        # TODO: may be better to instantiate an empty LynceusConfig, and use merge method here too.
        try:
            loaded_collection = json.loads(lynceus_config_dump)
        except JSONDecodeError as exc:
            # pylint: disable=raise-missing-from
            raise LynceusConfigError(f'Unable to JSON load config from specified string: "{lynceus_config_dump}".', from_exception=exc)

        if isinstance(loaded_collection, dict):
            return LynceusConfig(loaded_collection)

        return loaded_collection

    @staticmethod
    def load_json_params_from_lynceus_config(config_map: dict[str, any]):
        # Safe-guard: ensures specified parameter is a dict.
        if not isinstance(config_map, dict):
            raise LynceusConfigError(f'Specified parameter {config_map=} should be a dict ({type(config_map)} actually.')

        # Loads the specified config (e.g. custom user Project Profile metadata, jobs overriding/configuration options):
        #  - if key ends with the special json keyword, loads it
        #  - otherwise keep the key, value metadata pair unchanged
        loaded_lynceus_config: LynceusConfig = LynceusConfig()
        for config_param_with_json_key in list(config_map):
            # Checks if it is a json dump metadata.
            config_param_with_json_key_name: str = str(config_param_with_json_key)
            if not config_param_with_json_key_name.endswith(CONFIG_JSON_DUMP_KEY_END_KEYWORD):
                # It is NOT a json dump.

                # Checks if the value is a complex one.
                if isinstance(config_map[config_param_with_json_key_name], (dict, MutableMapping)):
                    # Calls recursively this method on the complex value.
                    loaded_lynceus_config[config_param_with_json_key_name].update(LynceusConfig.load_json_params_from_lynceus_config(config_map[config_param_with_json_key_name]).as_dict())
                else:
                    # Registers this single value.
                    LynceusConfig.recursive_merge(loaded_lynceus_config, {config_param_with_json_key_name: config_map[config_param_with_json_key_name]})

                continue

            # Loads the json value, and merges it recursively.
            loaded_metadata_key = config_param_with_json_key_name.removesuffix(CONFIG_JSON_DUMP_KEY_END_KEYWORD)
            value_dump = config_map[config_param_with_json_key]
            LynceusConfig.recursive_merge(loaded_lynceus_config, {loaded_metadata_key: LynceusConfig.load_from_config_str(value_dump)})

        return loaded_lynceus_config

    def save_partial_to_config_file(self, *, section_key_map: list[tuple], file_path: Path,
                                    key_transform_func=None,
                                    logger: Logger | None = None, log_prefix: str | None = ''):
        """
        Saves a copy of specified (section, key) tuple list, to file whose path is specified.

        :param section_key_map: a list of (section, key) tuple of configuration option to save (there is no error if any does not exist).
        :param file_path:
        :param key_transform_func: optional **key** transformation function (e.g. useful to merge project template LynceusMetadata configuration in the project to score with
                                    updated key, to state they belong to the template and not the project itself).
        :param logger:
        :param log_prefix:
        """
        config_parser: ConfigParser = ConfigParser()
        not_existing_value: str = 'no/th/ing'

        # Defines the metadata lynceus file, from complete configuration file, if defined.
        for metadata_option in section_key_map:
            section: str = metadata_option[0]
            key: str = metadata_option[1]

            # Attempts to retrive the value.
            value = self.get_config(section, key, default=not_existing_value)
            if value == not_existing_value:
                continue

            # Ensures section exists.
            if not config_parser.has_section(section):
                config_parser.add_section(section)

            # Transforms key if func is specified (useful to save project template config options, while processing the project to score).
            if key_transform_func:
                key = key_transform_func(key)

            # Registers it.
            if isinstance(value, Path):
                # - Path are saved as a string (not as a json dump, which is not loadable).
                value = str(value)
            elif not isinstance(value, str):
                # - all other type, but string, are saved as Json dump.
                key += CONFIG_JSON_DUMP_KEY_END_KEYWORD
                value = self.dump_to_config_str(value)
            #     - registers it inside config_parser.
            config_parser.set(section, key, value)

        if not config_parser.sections():
            logger.info(f'{log_prefix} no configuration option found to create Lynceus metadata file content among "{section_key_map}"')
            return

        if logger:
            logger.debug(f'{log_prefix} created Lynceus metadata file content:\n{LynceusConfig.dump_config_parser(config_parser)}')

        with open(file_path, 'w', encoding='utf8') as file:
            config_parser.write(file)

        if logger:
            logger.debug(f'{log_prefix} successfully written Lynceus metadata to file "{file_path}".')

    def is_empty(self):
        return len(self.__store) == 0

    def __repr__(self):
        return str(LynceusConfig.format_config(self))

    def __str__(self):
        return LynceusConfig.format_dict_to_string(LynceusConfig.format_config(self))

    def __or__(self, other):
        # Implements Python 3.9+ dict | operator allowing mixing between LynceusConfig and dict.
        return self.__store | other


class LynceusConfigJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, LynceusConfig):
            return dict(o)

        if isinstance(o, set):
            return list(o)

        if isinstance(o, Path):
            return str(o)

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)



class LynceusDict(dict):
    """
    Little dict enhancement allowing [.] attribute access, in addition to usual [key] ones.
    """
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as key_error:
            raise AttributeError from key_error

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError as key_error:
            raise AttributeError from key_error



from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.core.lynceus_client import LynceusClientClass
from .github_devops_analyzer import GithubDevOpsAnalyzer
from .gitlab_devops_analyzer import GitlabDevOpsAnalyzer
from ..core.config import (CONFIG_AUTHENTICATION_KEY,
                           CONFIG_PROJECT_CRED_SECRET,
                           CONFIG_PROJECT_CRED_SECRET_GITHUB_DEFAULT,
                           CONFIG_PROJECT_CRED_SECRET_GITLAB_DEFAULT,
                           CONFIG_PROJECT_KEY,
                           CONFIG_PROJECT_URI)
from ..core.config.lynceus_config import LynceusConfig


class DevOpsFactory(LynceusClientClass):
    def __init__(self, lynceus_session: LynceusSession, lynceus_exchange: LynceusExchange | None):
        super().__init__(lynceus_session, 'devops', lynceus_exchange)

    def new_devops_analyzer(self, uri: str, token: str):
        if 'gitlab' in uri:
            self._logger.debug(f'Instantiating a Gitlab implementation, according to uri \'{uri}\'')
            return GitlabDevOpsAnalyzer(self._lynceus_session, uri, token, self._lynceus_exchange)

        self._logger.debug(f'Instantiating a Github implementation, according to uri \'{uri}\'')
        return GithubDevOpsAnalyzer(self._lynceus_session, uri, token, self._lynceus_exchange)

    def get_access_token_simulating_anonymous_access(self, *, uri: str, lynceus_config: LynceusConfig):
        if 'gitlab' in uri:
            default_credential_secret_key: str = CONFIG_PROJECT_CRED_SECRET_GITLAB_DEFAULT
        else:
            default_credential_secret_key: str = CONFIG_PROJECT_CRED_SECRET_GITHUB_DEFAULT

        default_credential_secret: str = lynceus_config.get_config(CONFIG_AUTHENTICATION_KEY, default_credential_secret_key, default=None)
        if default_credential_secret:
            self._logger.warning(f'No specific "{CONFIG_PROJECT_CRED_SECRET}" in your [{CONFIG_AUTHENTICATION_KEY}] configuration,'
                                 f' using the configured "{default_credential_secret_key}" one.')
            lynceus_config[CONFIG_AUTHENTICATION_KEY][CONFIG_PROJECT_CRED_SECRET] = default_credential_secret
            return default_credential_secret

        error_message: str = f'Unable to find "{CONFIG_PROJECT_CRED_SECRET}" in your [{CONFIG_AUTHENTICATION_KEY}] configuration,'
        error_message += f' and there is no "{default_credential_secret_key}" configured. Update your configuration and try again.'

        self._logger.warning(error_message)
        raise ValueError(error_message)

    def check_and_enhance_configuration_for_anonymous_access(self, lynceus_config: LynceusConfig):
        uri: str = lynceus_config.get_config(CONFIG_PROJECT_KEY, CONFIG_PROJECT_URI)
        configured_credential_secret: str = lynceus_config.get_config(CONFIG_PROJECT_KEY, CONFIG_PROJECT_CRED_SECRET, default=None)
        if configured_credential_secret is None:
            lynceus_config[CONFIG_PROJECT_KEY][CONFIG_PROJECT_CRED_SECRET] = self.get_access_token_simulating_anonymous_access(uri=uri, lynceus_config=lynceus_config)



"""
This enum is mainly intended to be used by the client project to define
 which messages worth to be shown in UX.
"""

from enum import Enum


class LynceusMessageStatus(Enum):
    """
    Neutral informative message.
    """
    NEUTRAL = 'neutral'

    """
    Positive message leading to score increase.
    """
    POSITIVE = 'positive'

    """
    Positive message leading to score decrease.
    """
    NEGATIVE = 'negative'

    """
    Warning message.
    """
    WARNING = 'warning'

    """
    Error message.
    """
    ERROR = 'error'



from logging import Logger
from pathlib import Path

import s3fs

from lynceus.core.config import CONFIG_STORAGE_DYNAMIC_TYPE, CONFIG_STORAGE_IS_DYNAMIC, CONFIG_STORAGE_REMOTE_TYPE, LYNCEUS_S3_CONFIG_KEY
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.core.lynceus_client import LynceusClientClass
from lynceus.lynceus_exceptions import LynceusFileError


class S3Utils(LynceusClientClass):
    def __init__(self, *, lynceus_session: LynceusSession, lynceus_exchange: LynceusExchange | None, lynceus_s3_config: dict[str, str]):
        super().__init__(lynceus_session=lynceus_session, logger_name='s3', lynceus_exchange=lynceus_exchange)
        self.__lynceus_s3_config = lynceus_s3_config
        self.__s3filesystem = None

    def initialize(self):
        self._logger.info('Initializing s3fs according to configuration.')
        s3fs.S3FileSystem = S3FileSystemPatched

    def get_s3filesystem(self):
        if self.__s3filesystem is None:
            self.__s3filesystem = s3fs.S3FileSystem(**{LYNCEUS_S3_CONFIG_KEY: self.__lynceus_s3_config})

        return self.__s3filesystem

    def force_cache_refresh(self, path: Path = None):
        if self.__s3filesystem:
            self._logger.debug(f'Refreshing S3 fs cache ({path=}) ...')
            self.__s3filesystem.invalidate_cache(path=str(path) if path else None)

    def split_path(self, *, remote_file_path: str):
        return self.get_s3filesystem().split_path(remote_file_path)

    def list_remote_files(self, *, remote_root_path: Path,
                          recursive: bool, pattern: str | None = None,
                          maxdepth: int | None = None, withdirs: bool | None = None,
                          detail: bool = False):
        def _retrieve_remote_files():
            if recursive:
                # s3fs Globing feature does not support path with the 's3:/' prefix ...
                return self.get_s3filesystem().glob(str(Path(remote_root_path) / Path(pattern or '**/*')),
                                                    maxdepth=maxdepth, detail=detail)

            # Uses the find method, because ls one is NOT implemented in s3fs.
            # Some options are used only if pattern is NOT defined.
            find_kwargs = {
                'detail': detail
            }

            if pattern:
                find_kwargs.update({'prefix': pattern})
            else:
                find_kwargs.update(
                    {
                        'maxdepth': maxdepth or 1,
                        'withdirs': withdirs if not None else True
                    }
                )

            # s3 find feature does not support path with the 's3:/' prefix ...
            return self.get_s3filesystem().find(path=remote_root_path, **find_kwargs)

        def _extract_path(remote_file_path):
            return self.split_path(remote_file_path=remote_file_path)[1]

        try:
            self._logger.debug(f'Looking/globbing for remote files ({remote_root_path=}; {pattern=}) ...')
            all_remote_file_metadata = _retrieve_remote_files()

            # Processed result according to requested detail
            #  - only path are returned as a list
            if not detail:
                return [_extract_path(remote_file_path) for remote_file_path in all_remote_file_metadata]

            #  - path, and lots of metadata are returned as a dict
            return {_extract_path(remote_file_key): remote_file_detailed_metadata
                    for remote_file_key, remote_file_detailed_metadata in all_remote_file_metadata.items()}
        except Exception as exc:  # pylint: disable=broad-except
            # pylint: disable=raise-missing-from
            raise LynceusFileError(f'An error occured while looking/globbing for remote files ({remote_root_path=}; {pattern=}) ...', from_exception=exc)


# Patch and initialize s3fs once for all.
# pylint: disable=useless-super-delegation,abstract-method
class S3FileSystemPatched(s3fs.S3FileSystem):
    def __init__(self, *k, **kw):
        # Extracts s3 config key from parameters.
        self.lynceus_s3_config = kw.pop(LYNCEUS_S3_CONFIG_KEY, None)
        if not self.lynceus_s3_config:
            raise ValueError(f'{LYNCEUS_S3_CONFIG_KEY} is a mandatory parameter when initializing S3FileSystemPatched')

        lynceus: LynceusSession = LynceusSession.get_session(registration_key={'user': 's3filesystem'})
        logger: Logger = lynceus.get_logger('s3Init')
        logger.info(f'Using S3 config "{LynceusConfig.format_config(self.lynceus_s3_config)}".')

        # Builds client kwargs.
        client_kwargs = {
            'endpoint_url': self.lynceus_s3_config['s3_endpoint'],
        }

        # Adds any additional/extra parameters to client kwargs.
        client_kwargs.update(
            {key: value for key, value in self.lynceus_s3_config.items()
             if key not in (CONFIG_STORAGE_REMOTE_TYPE, 'endpoint_url', 'bucket_name', 'username',
                            'access_key_id', 'secret_access_key', 's3_endpoint', 'addressing_style',
                            CONFIG_STORAGE_IS_DYNAMIC, CONFIG_STORAGE_DYNAMIC_TYPE)}
        )
        logger.info(f'System will use these S3 client kwargs: "{LynceusConfig.format_config(client_kwargs)}".')

        super().__init__(*k,
                         key=self.lynceus_s3_config.get('access_key_id'),
                         secret=self.lynceus_s3_config.get('secret_access_key'),
                         client_kwargs=client_kwargs,
                         **kw)

    async def _rm(self, path, recursive=False, **kwargs):
        return await super()._rm(path, recursive, **kwargs)

    async def _rm_file(self, path, **kwargs):
        return await super()._rm_file(path, **kwargs)

    def created(self, path):
        return super().created(path)

    def cp_file(self, path1, path2, **kwargs):
        return super().cp_file(path1, path2, **kwargs)

    def sign(self, path, expiration=100, **kwargs):
        return super().sign(path, expiration, **kwargs)

    def ls(self, path, detail=True, **kwargs):
        # See: s3fs.core.S3FileSystem._find
        return self.find(path, maxdepth=1, withdirs=True, detail=detail, **kwargs)



import random
import re
import time
import timeit
from collections.abc import Iterable
from contextlib import contextmanager
from datetime import datetime, timezone
from logging import Logger
from pathlib import Path
from string import ascii_letters, digits

from setuptools import find_packages

# Default allowed characters when cleansing string values (e.g. activity/topic name).
from lynceus.core.config import DATETIME_FORMAT, DATETIME_FORMAT_SHORT

ALLOWED_CHARACTERS = ascii_letters + digits + 'ÀÂÄÆÇÈÉÊËÎÏÔŒÙÛÜàâäæçèéêëîïôœùûü'


def cleansed_str_value(value: str, *,
                       to_lower_case: bool = True,
                       replacement_character: str = '_',
                       allowed_characters: str = ALLOWED_CHARACTERS):
    if to_lower_case:
        value = value.lower()
    return ''.join(char if char in allowed_characters else replacement_character for char in value.strip())


def concatenate_string_with_limit(begin: str, extra: str, *, limit: int, truncate_begin: bool = True):
    if len(begin) >= limit:
        return begin[:limit]

    remaining_limit: int = limit - len(begin)
    kept_extra: str = extra
    if len(extra) > remaining_limit:
        if truncate_begin:
            kept_extra = extra[-remaining_limit:]
        else:
            kept_extra = extra[0:remaining_limit]

    return begin + kept_extra


def exec_and_return_time(func, /, *args) -> float:
    """
    Utility method allowing to measure execution time of specified function, and arguments.
    This function allows to easily change implementation at any time.

    IMPORTANT: does NOT support async function.

    :param func:
    :param args:
    :return:
    """
    return timeit.timeit(lambda: func(*args), number=1)


@contextmanager
def time_catcher() -> float:
    start = time.time()
    cpu_start = time.process_time()
    yield lambda: f'{(time.process_time() - cpu_start):0.03f} CPU seconds, {(time.time() - start):0.03f} elapsed seconds'


def parse_string_to_datetime(datetime_str: str, *,
                             datetime_format: str = DATETIME_FORMAT,
                             override_timezone: timezone | None = timezone.utc) -> datetime:
    try:
        final_datetime = datetime.strptime(datetime_str, datetime_format)
    except ValueError:
        final_datetime = datetime.strptime(datetime_str, DATETIME_FORMAT_SHORT)

    if override_timezone:
        final_datetime = final_datetime.replace(tzinfo=override_timezone)

    return final_datetime


def format_exception_human_readable(exc: Exception, *, quote_message: bool = False) -> str:
    result_begin: str = f'{exc.__class__.__name__}: '
    exc_msg: str = str(exc)

    return result_begin + (f'"{exc_msg}"' if quote_message else exc_msg)


def create_temporary_directory(root_dir: Path, seed: str | None = None):
    seed: str = seed or str(''.join(random.choice(ascii_letters) for _ in range(16)))
    tmp_dir: Path = root_dir / Path(seed)
    tmp_dir.mkdir(parents=True, exist_ok=True)
    return tmp_dir


def lookup_root_path(path_to_search_string: Path | str, remaining_iteration: int = 3, root_path: Path = Path().resolve()) -> Path:
    """
    Looks for the specified path_to_search_string starting from specified path (or current path, by default); checking
     iteratively in parent directory, during specified remaining_iteration iteration (default: 3).

    :param path_to_search_string:
    :param remaining_iteration:
    :param root_path:
    :return: the root path containing specified path_to_search_string (you need to concatenate result and path_to_search_string if you want a full path for it).
    """
    full_path: Path = root_path / Path(path_to_search_string)
    if full_path.exists():
        return root_path

    if not remaining_iteration:
        raise FileNotFoundError(f'Unable to find root_path of specified "{path_to_search_string}" path, after several iteration (last check in "{root_path}" directory).')

    return lookup_root_path(path_to_search_string, remaining_iteration - 1, root_path.parent)


def lookup_files_from_pattern(root_path: Path, pattern: str, *, min_file_size: float = None, case_insensitive: bool = True, logger: Logger = None):
    if case_insensitive:
        # Enhances the pattern to be case-insentitive.
        pattern = ''.join(map(lambda c: f'[{c.lower()}{c.upper()}]' if c.isalpha() else c, pattern))

    # Uses globbing in any case, according to the way pattern may have been enhanced to manage case.
    existing_files_list = list(root_path.glob(pattern))

    # Checks file size if needed.
    if min_file_size is not None:
        # Keeps only file whose size is greater of equal to specified size.
        existing_files_list = list(filter(lambda file: file.stat().st_size >= min_file_size, existing_files_list))

        if logger:
            logger.debug(f'After "file_size>={min_file_size} Bytes" filter, this is the list of files matching pattern "{pattern}": {existing_files_list=}')

    return existing_files_list


def check_file_exist_from_pattern(root_path: Path, pattern: str, *, min_file_size: float | None = None,
                                  case_insensitive: bool = True, logger: Logger | None = None):
    return len(lookup_files_from_pattern(root_path, pattern, min_file_size=min_file_size, case_insensitive=case_insensitive, logger=logger)) > 0


def lookup_available_packages(root_dir: Path | str, *, keep_children_packages: bool = False) -> set[str]:
    packages: list[str | bytes] = find_packages(root_dir)
    packages: set[str] = set(packages)
    if keep_children_packages:
        return packages

    # Removes all packages children.
    something_change: bool = True
    filtered_packages = packages
    while something_change:
        merged_children = set()
        for element in filtered_packages:
            # Merges all package children in the same set.
            merged_children ^= {child for child in filtered_packages - {element} if child.startswith(element)}

        # Updates filtered packages set if needed.
        filtered_packages -= merged_children

        # Registers something change.
        something_change = len(merged_children) > 0

    # Returns the final filtered packages set.
    return filtered_packages


def compute_file_line_count(file_path: Path):
    # Does not count one-line comment, empty line, line with only spaces characters, and docstring begin lines.
    # But following lines of docstring will unfortunately be counted, and it is an accepted limitation.
    source_code_line_pattern = re.compile(r'^\s*[^#"\s\']\S+.*$')
    source_code_line_count: int = 0
    with open(file_path, encoding='utf8') as file:
        for line in file:
            source_code_line_count += 1 if source_code_line_pattern.match(line) and len(line) > 4 else 0
    return source_code_line_count


def extract_class_fqn(specified_class: type) -> str:
    return f'{specified_class.__module__}.{specified_class.__name__}'


def dynamically_load_class(module_path: str, class_name: str):
    mod = __import__(module_path, fromlist=[class_name])
    return getattr(mod, class_name)


def inspect_attrs(obj, logger: Logger, patterns=None):
    pattern_info: str = 'with no condition' if not patterns else f'matching any of one of these patterns: {patterns}'
    logger.debug(f'Checking all Python attributes of instance "{obj}", {pattern_info}')
    for attr, value in obj.__dict__.items():
        if not patterns or any(pattern in attr for pattern in patterns):
            logger.debug(f'\t{attr=} => {value=}')


def flatten(collection):
    for item in collection:
        if isinstance(item, Iterable):
            yield from flatten(item)
        else:
            yield item


def filter_kwargs(*, args_filter: list[str], **kwargs):
    return dict(filter(lambda kv: print(f'{kv=}') or kv[0] in args_filter, kwargs.items()))



import logging.config
import logging.handlers
import sys
from logging import Logger
from pathlib import Path

from lynceus.core.config import CONFIG_GENERAL_KEY
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.lynceus_exceptions import LynceusConfigError
from lynceus.utils import format_exception_human_readable, lookup_root_path


class LynceusSession:
    DEFAULT_SALT: str = 'lynceus'

    # Additional even more verbose than DEBUG log level.
    TRACE = 5

    __internal_key_preventing_external_usage_of_init = object()

    __REGISTERED_SESSIONS: dict = {}

    def __init__(self, *, salt: str = DEFAULT_SALT, root_logger_name: str = 'Ubeeko', load_core_default_config: bool = True,
                 _creation_key=None):
        # Safe-guard: prevents direct call to this method.
        if _creation_key != LynceusSession.__internal_key_preventing_external_usage_of_init:
            raise RuntimeError(f'You should always use LynceusSession.get_session() method when requesting a LynceusSession ({salt=}).')

        # Initializes internal Lynceus config.
        self.__config = LynceusConfig()
        self.__root_logger_name = root_logger_name

        # Loads configuration file(s):
        #  - first the default config according to load_core_default_config toggle
        #  - then the default and user config corresponding to specify salt, if not the default one
        loaded_config_files = self.__load_configuration(salt, load_core_default_config=load_core_default_config)

        # Additional even more verbose than DEBUG log level.
        logging.TRACE = LynceusSession.TRACE
        logging.addLevelName(LynceusSession.TRACE, 'TRACE')

        # Gets default/root Lynceus logger.
        self.__logger = self.get_logger()
        root_level: int = logging.getLevelName(self.get_config(CONFIG_GENERAL_KEY, 'logger.root.level', default='INFO'))
        self.__logger.setLevel(root_level)

        # Configures main root logger.
        self.__configure_logger(self.__logger)

        # Disables default root logger, removing all default handlers on it.
        # This instruction is not enough, because it automatically created another handler: logging.basicConfig(force=True)
        logging.root.propagate = False
        for default_root_handler in logging.root.handlers:
            logging.root.removeHandler(default_root_handler)

        # Informs now that Logger is initialized.
        self.__logger.info(f'This is the loading information of looked up configuration files (Format=<Path: isLoaded?>) "{loaded_config_files}".')

    @staticmethod
    # pylint: disable=protected-access
    def get_session(*, salt: str = DEFAULT_SALT, registration_key: dict[str, str] | None = None,
                    root_logger_name: str = 'Lynceus', load_core_default_config: bool = True):
        # Safe-guard: a registration key is highly recommended if not an internal session (which would be the case with salt == DEFAULT_SALT).
        if registration_key is None:
            if salt != LynceusSession.DEFAULT_SALT:
                raise LynceusConfigError('It is mandatory to provide your own registration_key when getting a session.' +
                                         f' System will automatically consider it as {registration_key}, but you may not be able to retrieve your session.' +
                                         ' And it may lead to memory leak.')
            registration_key = {'default': salt}

        # Turns the registration_key to an hashable version.
        if registration_key is not None:
            registration_key = frozenset(registration_key.items())

        # Creates and registers the session if needed.
        if registration_key in LynceusSession.__REGISTERED_SESSIONS:
            session: LynceusSession = LynceusSession.__REGISTERED_SESSIONS[registration_key]
            session.__logger.debug(f'Successfully retrieved cached Session for registration key "{registration_key}": {session}.')
        else:
            session: LynceusSession = LynceusSession(salt=salt, root_logger_name=root_logger_name,
                                                     load_core_default_config=load_core_default_config,
                                                     _creation_key=LynceusSession.__internal_key_preventing_external_usage_of_init)

            LynceusSession.__REGISTERED_SESSIONS[registration_key] = session
            session.__logger.debug(f'Successfully registered new Session (new total count={len(LynceusSession.__REGISTERED_SESSIONS)})'
                                   f' for registration key "{registration_key}": {session}.'
                                   f' This is its loaded configuration (salt={salt}):\n' +
                                   f'{LynceusConfig.format_dict_to_string(LynceusConfig.format_config(session.__config), indentation_level=1)}')

        # TODO: implement another method allowing to free a specific registered session
        # TODO: turn LynceusSession as a Context, allowing usage with the 'with' keyword to automatic free the session once used

        # Returns the Lynceus session.
        return session

    def __load_configuration(self, salt: str, load_core_default_config: bool) -> dict[Path, bool]:
        # Defines the potential configuration file to load.
        config_file_meta_list = []
        if load_core_default_config:
            # Adds the Lynceus default configuration file as requested.
            config_file_meta_list.append({'config_path': f'misc/{self.DEFAULT_SALT}.default.conf', 'root_path': Path(__file__).parent})

        # Adds default configuration file corresponding to specified salt, if not the default one.
        if salt != self.DEFAULT_SALT:
            config_file_meta_list.append({'config_path': f'misc/{salt}.default.conf'})

        # Adds user configuration file corresponding to specified salt.
        config_file_meta_list.append({'config_path': f'{salt}.conf'})

        # Looks up for configuration file(s), starting with a "default" one, and then optional user one.
        loaded_config_files: dict[Path, bool] = {}
        for config_file_meta in config_file_meta_list:
            conf_file_name = config_file_meta['config_path']
            conf_file_root_path = config_file_meta.get('root_path', Path().resolve())

            loaded: bool = False
            relative_file: Path = Path(conf_file_name)
            try:
                full_path: Path = lookup_root_path(relative_file, root_path=conf_file_root_path) / relative_file

                # Merges it in internal Lynceus config.
                self.__config.update_from_configuration_file(full_path)

                loaded = True
            except FileNotFoundError as exc:
                # Since recent version, there is no more mandatory configuration file to have, and several files can be loaded.
                # N.B.: it is valid, there is no logger yet here, so using print ...
                self.get_logger('internal').warning(f'WARNING: Unable to load "{conf_file_name}" configuration file,'
                                                    f' from "{conf_file_root_path}"'
                                                    f' (not existing configuration file is valid) =>'
                                                    f' {format_exception_human_readable(exc, quote_message=True)}')

            # Registers configuration file loading information.
            loaded_config_files[relative_file] = loaded

        # Returns configuration file loading mapping, for information.
        return loaded_config_files

    def has_config_section(self, section: str):
        return self.__config.has_section(section)

    def get_config(self, section: str, key: str, *, default: object = LynceusConfig.UNDEFINED_VALUE):
        return self.__config.get_config(section, key, default=default)

    def is_bool_config_enabled(self, section: str, key: str):
        return self.__config.is_bool_config_enabled(section=section, key=key)

    def get_config_section(self, section: str):
        return self.__config[section]

    def get_lynceus_config_copy(self) -> LynceusConfig:
        """
        Returns a copy of the internal Lynceus config, can be very interesting to get default config.
        :return: a copy of the internal Lynceus config, can be very interesting to get default config.
        """
        return self.__config.copy()

    @staticmethod
    def __create_stream_handler(stream, formatter, level, message_filter=None):
        handler = logging.StreamHandler(stream=stream)
        handler.setLevel(level)
        handler.setFormatter(formatter)
        if message_filter:
            handler.addFilter(message_filter)
        return handler

    def __configure_logger(self, logger: Logger):
        if not logger.handlers:
            if self.is_bool_config_enabled(CONFIG_GENERAL_KEY, 'logger.console.enable'):
                message_format: str = self.get_config(CONFIG_GENERAL_KEY, 'logger.console.format',
                                                      default='{asctime:20} {name:16} {levelname:8} {message}')
                date_format: str = self.get_config(CONFIG_GENERAL_KEY, 'logger.console.date_format', default='%Y/%m/%d %H:%M:%S')
                level: int = logging.getLevelName(self.get_config(CONFIG_GENERAL_KEY, 'logger.console.level', default='INFO'))
                formatter = logging.Formatter(message_format, date_format, '{')

                # Configures error output (from Warning levels).
                error_output_handler = LynceusSession.__create_stream_handler(sys.stderr, formatter,
                                                                              max(level, logging.WARNING))
                logger.addHandler(error_output_handler)

                # Configures standard output (from configured Level, if lower than Warning,
                #  and excluding everything from Warning and higher).
                if level < logging.WARNING:
                    # pylint: disable=unnecessary-lambda-assignment
                    standard_output_filter = lambda record: record.levelno < logging.WARNING
                    standard_output_handler = LynceusSession.__create_stream_handler(sys.stdout, formatter, level,
                                                                                     standard_output_filter)
                    logger.addHandler(standard_output_handler)

            if self.is_bool_config_enabled(CONFIG_GENERAL_KEY, 'logger.file.enable'):
                message_format: str = self.get_config(CONFIG_GENERAL_KEY, 'logger.file.format',
                                                      default='{asctime:20} {name:16} {levelname:8} {message}')
                date_format: str = self.get_config(CONFIG_GENERAL_KEY, 'logger.file.date_format', default='%Y/%m/%d %H:%M:%S')
                level: int = logging.getLevelName(self.get_config(CONFIG_GENERAL_KEY, 'logger.file.level', default='INFO'))
                output_file: str = self.get_config(CONFIG_GENERAL_KEY, 'logger.file.path', default=None)

                if output_file is None:
                    raise LynceusConfigError('You must at least specify "logger.file.path" if you enable "logger.file.enable".')

                # Manages file open mode (append by default, override if configured).
                open_mode: str = 'w' if self.is_bool_config_enabled(CONFIG_GENERAL_KEY, 'logger.file.override') else 'a'
                handler = logging.FileHandler(output_file, mode=open_mode)
                formatter = logging.Formatter(message_format, date_format, '{')
                handler.setLevel(level)
                handler.setFormatter(formatter)
                logger.addHandler(handler)

    def get_logger(self, name: str = None) -> Logger:
        complete_name: str = f'{self.__root_logger_name}.{name}' if name else self.__root_logger_name
        return logging.getLogger(complete_name)



from enum import Enum


class FileMetadataType(Enum):
    DATASET: str = 'Dataset'
    DATASET_EXCERPT: str = 'Data excerpt'
    DATASET_FULL: str = 'Full dataset'

    TEST_SET_INTERIM: str = 'Interim Test Set'
    TEST_SET_FINAL: str = 'Final Test Set'

    DATA_DICTIONARY: str = 'Data dictionary'

    ASSIGNMENT_SOLUTION: str = 'Solution File'
    ASSIGNMENT_SAMPLE: str = 'Template Solution File'
    ASSIGNMENT_BENCHMARK: str = 'Benchmark Solution File'
    ASSIGNMENT_SUBMISSION: str = 'Submission File'

    CUSTOM_SCORER: str = 'Custom scorer'

    OTHER: str = 'Other'
    UNSPECIFIED: str = 'Unspecified'



import shutil
from abc import ABCMeta
from logging import Logger
from pathlib import Path
from typing import Generic, TypeVar

import pandas as pd
import pyarrow
import pyarrow.parquet as pq
from fsspec.asyn import AsyncFileSystem
from pandas import DataFrame
from pandas.io.common import _get_filepath_or_buffer

from lynceus.core.config import CONFIG_STORAGE_LOCAL, LYNCEUS_S3_CONFIG_KEY
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.files.remote.s3 import S3FileSystemPatched, S3Utils
from lynceus.lynceus_exceptions import LynceusFileError

# pylint: disable=invalid-name
FileSystemType = TypeVar("FileSystemType", bound=AsyncFileSystem)


class LynceusFile(Generic[FileSystemType], metaclass=ABCMeta):
    S3_PATH_BEGIN = 's3://'

    FILE_STORAGE_PATH_SEPARATOR: str = '|'

    def __init__(self,
                 path: Path,
                 logger: Logger,
                 filesystem: FileSystemType = None):
        self._path: Path = path
        self._logger: Logger = logger
        self._filesystem: FileSystemType = filesystem

    @staticmethod
    def extract_storage_and_path(file_metadata: str):
        # Checks if there is storage information in the metadata.
        if LynceusFile.FILE_STORAGE_PATH_SEPARATOR in file_metadata:
            file_metadata_parts = file_metadata.split(LynceusFile.FILE_STORAGE_PATH_SEPARATOR)
            return file_metadata_parts[0], file_metadata_parts[1]

        # There is none, so consider it as a file hosted on Local storage.
        return CONFIG_STORAGE_LOCAL, file_metadata

    @staticmethod
    def build_file_metadata(storage_name: str, file_path: str):
        return f'{storage_name}{LynceusFile.FILE_STORAGE_PATH_SEPARATOR}{file_path}'

    def read_parquet(self, **params) -> DataFrame:
        """
        Reads corresponding (local or remote) file, considering it as a parquet file, with optional parameters.
        :param params: optional parameters (usually, it can be columns, to specify which columns
                        to read from parquet file).
        :return: corresponding DataFrame.
        """
        # Cf. https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_parquet.html
        # Cf. http://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html

        self._logger.debug(f'Reading file {self} ...')
        # pylint: disable=c-extension-no-member
        try:
            return pd.read_parquet(self.get_path(), storage_options=self.get_storage_options(), **params)
        except pyarrow.lib.ArrowInvalid as exc:
            # Special (somehow dirty) hack allowing to get a workaround about some issue with invalid datetime.
            # Initially needed for Exco project.
            if 'would result in out of bounds timestamp' not in str(exc):
                # N.B. it is not the "issue" for which the workaround has been developed.
                raise

            self._logger.warning(f'Caught "out of bounds timestamp" error while reading file {self} ... trying to read it with unsafe method as a workaround.')

            # The key of the workaround is the use the **safe=False** option on the to_pandas method,
            #  which can't be specified through Pandas.
            # But, because we are breaking down here the pd.read_parquet method, we need to work on the path
            #  using the pandas.io.common._get_filepath_or_buffer method to create a remote path if needed.
            ioargs = _get_filepath_or_buffer(self.get_path(), storage_options=self.get_storage_options())
            parquet_table = pq.read_table(ioargs.filepath_or_buffer, **params)
            return parquet_table.to_pandas(safe=False)

    def write_to_parquet(self, dataframe: DataFrame, **kwargs):
        # N.B.: **kwargs is the opportunity to provide parameters for internal implementation (e.g. PyArrow),
        #  for instances, pyarrow filters param, to better control what is load in memory.
        #
        # Cf. https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_parquet.html
        # Cf. https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#io-parquet
        # Cf. https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_table.html

        self._logger.debug(f'Writing specified DataFrame to file {self} ...')
        dataframe.to_parquet(self.get_path(),
                             coerce_timestamps='ms', allow_truncated_timestamps=True,
                             storage_options=self.get_storage_options(),
                             **kwargs)

    def get_storage_options(self):
        # pylint: disable=no-self-use
        return None

    def is_local(self):
        raise NotImplementedError()

    def is_remote(self):
        return not self.is_local()

    def delete(self):
        self._logger.debug(f'Deleting file {self} ...')
        return self._do_delete()

    def _do_delete(self):
        raise NotImplementedError()

    def download_to(self, destination: Path, *, create_sub_directories: bool = True):
        self._logger.debug(f'Retrieving/downloading file to "{destination}" from {self} ...')
        return self._do_download_to(destination=destination, create_sub_directories=create_sub_directories)

    def _do_download_to(self, *, destination: Path, create_sub_directories: bool):
        raise NotImplementedError()

    def exists(self, *, reason: str = None):
        check_msg: str = f'Checking existence of file {self}'
        if reason:
            check_msg += f' (reason: {reason})'
        self._logger.debug(f'{check_msg} ...')
        return self._do_exists()

    def _do_exists(self):
        raise NotImplementedError()

    def list_files(self, *, recursive: bool = False, pattern: str | None = None, **kwargs):
        self._logger.debug(f'Listing files from {self}, {pattern=} ...')
        return self._do_list_files(recursive=recursive, pattern=pattern, **kwargs)

    def _do_list_files(self, *, recursive: bool, pattern: str | None = None, **kwargs):
        raise NotImplementedError()

    def copy_to(self, destination: Path, *, create_sub_directories: bool = True):
        self._logger.debug(f'Copying {self} to local file {destination} ...')
        return self._do_copy_to(destination=destination, create_sub_directories=create_sub_directories)

    def _do_copy_to(self, *, destination: Path, create_sub_directories: bool):
        raise NotImplementedError()

    def get_name(self) -> str:
        return self._path.name

    @property
    def path(self) -> Path:
        return self._path

    def get_path(self) -> str:
        return str(self._path)

    def get_raw_path(self) -> str:
        raise NotImplementedError()

    def get_relative_path(self) -> str:
        """
        :return: for remote file: the relative path from remote storage container, for local file: same than raw_path.
        """
        raise NotImplementedError()

    def get_parent_path(self) -> Path:
        return self._path.parent

    def parent_exists(self):
        self._logger.debug(f'Checking existence of parent folder of file {self} ...')
        return self._do_parent_exists()

    def _do_parent_exists(self):
        raise NotImplementedError()

    def get_extension(self) -> str:
        return self._path.suffix

    def __str__(self):
        return f'"{self.__class__.__name__}" with path "{self._path}"'

    def __repr__(self):
        return str(self)


class _LocalLynceusFile(LynceusFile[AsyncFileSystem]):
    def is_local(self):
        return True

    def _do_delete(self):
        self._path.unlink()

    def _do_download_to(self, *, destination: Path, create_sub_directories: bool):
        return self._do_copy_to(destination=destination, create_sub_directories=create_sub_directories)

    def _do_exists(self):
        return self._path.exists()

    def _do_parent_exists(self):
        return self.get_parent_path().exists()

    def _do_list_files(self, *, recursive: bool, pattern: str | None = None, **kwargs):
        if not recursive:
            return self._path.iterdir()
        return self._path.glob(pattern or '**/*')

    def _do_copy_to(self, *, destination: Path, create_sub_directories: bool):
        if not destination.parent.exists():
            if create_sub_directories:
                destination.parent.mkdir(parents=True, exist_ok=True)
            else:
                raise LynceusFileError(f'Parent directory of specified destination "{destination}" does not exist;' +
                                       ' you should either create it yourself, or use the corresponding option.')

        # Requests the copy.
        shutil.copyfile(self.get_path(), destination)

    def get_raw_path(self) -> str:
        return str(self._path)

    def get_relative_path(self) -> str:
        return self.get_raw_path()


class _RemoteS3LynceusFile(LynceusFile[S3FileSystemPatched]):
    # In addition there is a self.S3_PATH_BEGIN usage in Factory which should be adapted (in case it is NOT S3 !).
    def __init__(self, path: Path, logger: Logger, s3filesystem: S3FileSystemPatched, s3_utils: S3Utils):
        super().__init__(path, logger, filesystem=s3filesystem)
        self.__s3_utils = s3_utils

    def get_storage_options(self):
        # N.B.: in our Patched remote fs System, we added the needed lynceus_s3_config.
        storage_options = {
            'anon': False,
            LYNCEUS_S3_CONFIG_KEY: self._filesystem.lynceus_s3_config
        }

        # Checks if it is an OVH remote storage.
        if '.ovh.' in self._filesystem.lynceus_s3_config['s3_endpoint']:
            # Hacks ACL information to workaround OVH Bug, with default ACL specified by s3fs/botocore.
            # Leading to an useless "OSError: [Errno 22] Invalid Argument." ...
            storage_options.update(
                {
                    's3_additional_kwargs': {'ACL': 'private'}
                }
            )

        return storage_options

    def is_local(self):
        return False

    def _do_delete(self):
        # TODO: fix this remote file delete method ...
        self._filesystem.delete(self.get_raw_path())

    # pylint: disable=unused-argument
    def _do_download_to(self, *, destination: Path, create_sub_directories: bool):
        return self._filesystem.get(self.get_path(), str(destination))

    def _do_exists(self):
        # N.B.: in case exists check does not work properly, it is possible to use ls, forcing cache refresh.
        # bucket, key, version_id = self.__s3filesystem.split_path(self._path)
        # Checks with a forced refresh, in case the file has been generated by this Data Exporter execution.
        # return len(self.__s3filesystem.ls(key, refresh=True)) > 0

        # Checks if it exists, letting internal s3 FS to use its cache system.
        # pylint: disable=broad-except
        try:
            # TODO: understand the Bad request which happens sometimes ...
            return self._filesystem.exists(self.get_raw_path())
        except Exception as exc:
            self._logger.warning(f'An exception occured while checking existence of remote file "{self}" (it will be considered as not existing).', exc_info=exc)
            # Can happen for instance with a bad/broken specified path;
            #   caller raises an exception on none existence which is enough (so here, we just return the file does not exist).
            return False

    def _do_parent_exists(self):
        return self._filesystem.exists(self.get_raw_path_from_remote_path(self.get_parent_path()))

    # pylint: disable=arguments-differ
    def _do_list_files(self, *, recursive: bool, pattern: str | None = None,
                       maxdepth: int | None = None, withdirs: bool | None = None,
                       detail: bool = False):
        return self.__s3_utils.list_remote_files(remote_root_path=Path(self.get_raw_path()),
                                                 recursive=recursive,
                                                 pattern=pattern,
                                                 maxdepth=maxdepth,
                                                 withdirs=withdirs,
                                                 detail=detail)

    # pylint: disable=unused-argument
    def _do_copy_to(self, *, destination: Path, create_sub_directories: bool):
        if self.is_remote() and destination.is_absolute():
            raise LynceusFileError(f'You should use only relative Path with remote file ("{self}"), which is not the case of destination "{destination}"')

        bucket_name, _, _ = self.__s3_utils.split_path(remote_file_path=self.get_path())
        return self._filesystem.copy(self.get_path(), str(Path(bucket_name) / destination))

    def get_path(self) -> str:
        # Important: to work, we must ensure the S3 PATH Begin is unaltered here (the double slash is mandatory ...).
        return LynceusFile.S3_PATH_BEGIN + self.get_raw_path()

    @staticmethod
    def get_raw_path_from_remote_path(path: Path):
        # Removes the 's3:/' prefix to get a raw path.
        raw_path_from_remote_path = str(path)[len(LynceusFile.S3_PATH_BEGIN) - 1:]

        # Safe-guard: ensures there is at least one '/' in the final raw path (which is NOT the case for remote 'root path',
        #  to avoid issue with s3fs path splitting feature, and avoid 'Could not traverse all s3' issue).
        if '/' not in raw_path_from_remote_path:
            raw_path_from_remote_path += '/'

        return raw_path_from_remote_path

    def get_raw_path(self) -> str:
        return self.get_raw_path_from_remote_path(self._path)

    def get_relative_path(self) -> str:
        _, rpath, _ = self.__s3_utils.split_path(remote_file_path=self.get_path())
        return rpath

    def __str__(self):
        return f'"{self.__class__.__name__}" with path "{self._path}" on remote "{LynceusConfig.format_config(self._filesystem.lynceus_s3_config)}"'



from logging import Logger
from pathlib import Path

from lynceus.core.config import (CONFIG_PROJECT_DYNAMIC_STORAGE_MANDATORY_PARAM_MAP,
                                 CONFIG_PROJECT_KEY,
                                 CONFIG_PROJECT_ROOT_PATH_HOLDER,
                                 CONFIG_STORAGE_LOCAL)
from lynceus.core.config.lynceus_config import LynceusConfig
from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.files.file_factory import LynceusFileFactory
from lynceus.files.lynceus_file import LynceusFile
from lynceus.lynceus_exceptions import LynceusConfigError, LynceusFileError
from lynceus.utils import lookup_root_path


def create_storage_file_factory(*, name: str,
                                lynceus_session: LynceusSession,
                                lynceus_config: LynceusConfig,
                                logger: Logger,
                                log_prefix: str,
                                remote_config_section: str | None,
                                remote_mode_forced_by_cli: bool = True,
                                source_path_format: str = '{target}',
                                dest_path_format: str = '{dest_file_name}',
                                lynceus_exchange: LynceusExchange | None = None) -> LynceusFileFactory | None:
    # Safe-guard: ensure specified config_section exists in Lynceus configuration.
    if remote_config_section and not lynceus_config.has_section(remote_config_section):
        logger.warning(f'{log_prefix} unable to register storage "{name}", because configuration section "{remote_config_section}" does not exist. Fix your configuration.')
        return None

    # Creates a new Lynceus file factory corresponding to needs.
    return LynceusFileFactory(
        lynceus_session=lynceus_session,
        lynceus_config=lynceus_config,
        remote_config_section=remote_config_section,
        remote_mode_forced_by_cli=remote_mode_forced_by_cli,
        source_path_format=source_path_format,
        dest_path_format=dest_path_format,
        lynceus_exchange=lynceus_exchange,
    )


def extract_dynamic_remote_storage_params(lynceus_config: LynceusConfig) -> dict[str, str | int]:
    if not lynceus_config.has_section(CONFIG_PROJECT_KEY):
        raise LynceusConfigError(f'Unable to find [{CONFIG_PROJECT_KEY}] configuration section in specified configuration file.')

    dynamic_remote_storage_params = {}
    for param, is_mandatory in CONFIG_PROJECT_DYNAMIC_STORAGE_MANDATORY_PARAM_MAP.items():
        value = lynceus_config.get_config(CONFIG_PROJECT_KEY, param, default=None)
        if value is None:
            if is_mandatory:
                raise LynceusConfigError(f'Unable to find "{param}" option (mandatory for dynamic remote storage) inside' +
                                         f' [{CONFIG_PROJECT_KEY}] configuration section in specified configuration file.')
            continue

        # Checks the type of value, can be either:
        #  - string if coming from a static configuration file
        #  - int if coming from API, CLI or Tests
        if isinstance(value, str):
            value = value if not value.isnumeric() else int(value)

        dynamic_remote_storage_params[param] = value

    return dynamic_remote_storage_params


def get_lynceus_file_from_metadata(*,
                                   file_metadata: str,
                                   lynceus_config: LynceusConfig,
                                   logger: Logger,
                                   log_prefix: str,
                                   storage_file_factory_map: dict[str, LynceusFileFactory],
                                   locally_retrieved_repository_root_path: Path | None,
                                   must_exist: bool,
                                   overriden_root_path_if_local: Path = None) -> LynceusFile:
    # Extracts storage name and file path from metadata.
    storage_name, file_path = LynceusFile.extract_storage_and_path(file_metadata)
    dest_path_format: str | None = None
    override_dest_path_kwargs: dict[str, str] | None = None

    storage_file_factory: LynceusFileFactory = storage_file_factory_map.get(storage_name)
    if not storage_file_factory:
        raise LynceusConfigError(f'{log_prefix} file option ("{file_metadata}"), is hosted on remote storage "{storage_name}"' +
                                 ', which is not configured!' +
                                 f' Available/configured remote storages: {storage_file_factory_map}.')

    # Manages dynamic remote storage if needed.
    if storage_file_factory.is_dynamic_remote:
        # TODO: limitation is that ALL remote file coming from a dynamic remote, share the same parameters linked to the parent project.
        # Thus: atm it is NOT possible to have a reference file on a dynamic remote and the solution file on another dynamic remote, for the same project.
        try:
            dynamic_remote_storage_params: dict[str, str | int] = extract_dynamic_remote_storage_params(lynceus_config)
            storage_file_factory.update_dynamic_storage_params(dynamic_remote_storage_params)

            # Checks if the path must be formatted.
            if '{' in file_path:
                dest_path_format = str(file_path)
                override_dest_path_kwargs = dynamic_remote_storage_params

        except LynceusConfigError as exc:
            # pylint: disable=raise-missing-from
            raise LynceusFileError(f'Unable to prepare system to use dynamic remote storage "{storage_name}"', exc)

    # Manages Local file if needed:
    if storage_name == CONFIG_STORAGE_LOCAL:
        if overriden_root_path_if_local:
            file_path: str = str(lookup_root_path(file_path, remaining_iteration=4, root_path=overriden_root_path_if_local) / Path(file_path))
        else:
            #  - adds special CONFIG_PROJECT_ROOT_PATH_HOLDER keyword at beginning if relative path
            if CONFIG_PROJECT_ROOT_PATH_HOLDER not in file_path and not file_path.startswith('/'):
                file_path: str = CONFIG_PROJECT_ROOT_PATH_HOLDER + file_path

        #  - replaces CONFIG_PROJECT_ROOT_PATH_HOLDER keyword by retrieved repository root path
        if CONFIG_PROJECT_ROOT_PATH_HOLDER in file_path:
            if not locally_retrieved_repository_root_path:
                raise ValueError('Repository should have been locally retrieved for this request.')

            root_dir: str = str(locally_retrieved_repository_root_path)
            file_path: str = file_path.replace(CONFIG_PROJECT_ROOT_PATH_HOLDER, root_dir + '/')

    # Creates a LynceusFile instance.
    logger.debug(f'{log_prefix} creating LynceusFile ({file_path=}; {dest_path_format=}; {override_dest_path_kwargs=}) ...')
    lynceus_file: LynceusFile = storage_file_factory.new_file(source_name=None, source_file_name=file_path,
                                                              dest_path_format=dest_path_format,
                                                              override_dest_path_kwargs=override_dest_path_kwargs,
                                                              create_sub_directories=False,
                                                              must_exist=must_exist)

    return lynceus_file


def retrieve_remote_file_locally(*, lynceus_file: LynceusFile,
                                 logger: Logger,
                                 log_prefix: str,
                                 dest_dir_path: Path,
                                 extension_if_none: str | None = None) -> Path | None:
    # This method is useful in several situations, for instance:
    #  - scoring engine is unable to work with remote file ... retrieve the remote file locally first
    #  - third-party tool configuration file can be overriden and put on remote storage, so we retrieve them locally first
    if lynceus_file is None:
        return None

    if lynceus_file.is_local():
        return lynceus_file.path

    local_file_name: str = lynceus_file.get_name()
    if not lynceus_file.get_extension() and extension_if_none:
        local_file_name += extension_if_none

    local_path: Path = dest_dir_path / Path(local_file_name)
    lynceus_file.download_to(local_path)
    logger.debug(f'{log_prefix} saved {lynceus_file=} to {local_path=} to ease usage.')
    return local_path



from lynceus.utils import format_exception_human_readable


class LynceusError(Exception):
    """Base Lynceus exception, all specific exceptions inherits from it."""

    def __init__(self, message, from_exception: Exception | None = None):
        super().__init__()
        self.__message = message
        if from_exception:
            self.__message += f'; caused by {format_exception_human_readable(from_exception, quote_message=True)}.'

    def __str__(self):
        return self.__message


class LynceusConfigError(LynceusError):
    pass


class LynceusJobError(LynceusError):
    pass


class LynceusJobCancelExecutionError(LynceusJobError):
    pass


class LynceusFileError(LynceusError):
    pass



from asyncio import Queue
from uuid import UUID, uuid4


class LynceusExchange(Queue):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__uid: UUID = uuid4()

    def __str__(self):
        return f'LynceusExchange-{self.__uid}'



import time
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from pathlib import Path

import gitlab
from gitlab.const import AccessLevel

from lynceus.core.config import DATETIME_FORMAT_SHORT
from lynceus.core.exchange.lynceus_exchange import LynceusExchange
from lynceus.core.lynceus import LynceusSession
from lynceus.devops.devops_analyzer import DevOpsAnalyzer
from lynceus.utils import flatten, parse_string_to_datetime
from lynceus.utils.lynceus_dict import LynceusDict


def gitlab_exception_handler(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except gitlab.exceptions.GitlabError as error:
            # Intercepts permission error.
            if error.response_code in (401, 403):
                raise PermissionError(
                    'You don\'t have enough permission to perform this operation on Gitlab.') from error
            if error.response_code == 404:
                raise NameError('Unable to find requested Object.') from error

            # Raises any other error.
            raise

    return func_wrapper


# Cf. https://python-gitlab.readthedocs.io/en/stable/api-usage.html
# Cf. https://docs.gitlab.com/ee/api/README.html
# Cf. https://docs.gitlab.com/ee/api/api_resources.html
class GitlabDevOpsAnalyzer(DevOpsAnalyzer):
    # Cf. https://docs.gitlab.com/ce/api/project_import_export.html#export-status
    # Cf. https://docs.gitlab.com/ce/api/project_import_export.html#import-status
    IMPORT_EXPORT_STATUS_SUCCESS: str = 'finished'
    IMPORT_EXPORT_STATUS_FAILED: str = 'failed'
    IMPORT_EXPORT_STATUS_NONE: str = 'none'

    def __init__(self, lynceus_session: LynceusSession, uri: str, token: str, lynceus_exchange: LynceusExchange):
        super().__init__(lynceus_session, uri, 'gitlab', lynceus_exchange)
        self.__manager = gitlab.Gitlab(uri, private_token=token)
        self.__current_token: dict | None = None

    # The following methods are only for uniformity and coherence.
    def _extract_user_info(self, user) -> LynceusDict:
        # Extracts information which are always available.
        user_info = {
            'id': user.id,
            'name': user.name,
            'login': user.username,
            'username': user.username
        }

        # Extracts extra information if available.
        for extra_info_key, extra_info_attr in ('e-mail', 'public_email'), ('avatar_url', 'avatar_url'), ('bio', 'bio'):
            value = user.attributes[extra_info_attr] if extra_info_attr in user.attributes else self.INFO_UNDEFINED
            user_info.update({extra_info_key: value})

        return LynceusDict(user_info)

    def _extract_group_info(self, group) -> LynceusDict:
        return LynceusDict({
            'id': group.id,
            'name': group.name,
            'path': group.full_path
        })

    def _extract_project_info(self, project) -> LynceusDict:
        return LynceusDict({
            'id': project.id,
            'name': project.name,
            'path': project.path_with_namespace,
            'web_url': project.web_url
        })

    def _extract_member_info(self, member) -> LynceusDict:
        member_info = {
            'id': member.id,
            'name': member.name,
            'login': member.username,
            'username': member.username,
            'state': member.state
        }

        # Extracts extra information if available.
        for extra_info_key, extra_info_attr in ('parent_id', 'group_id'), ('parent_id', 'project_id'):
            value = member.attributes[extra_info_attr] if extra_info_attr in member.attributes else None
            if value:
                member_info.update({extra_info_key: value})

        return LynceusDict(member_info)

    def _extract_issue_event_info(self, issue_event, **kwargs) -> LynceusDict:
        return LynceusDict({
                               'id': issue_event.id,
                               'issue_id': issue_event.target_iid,
                               'action': issue_event.action_name,
                               'target_type': issue_event.target_type,
                               'created_at': parse_string_to_datetime(datetime_str=issue_event.created_at),
                               'author': issue_event.author['name'],
                               'title': issue_event.target_title,
                               'issue_web_url': kwargs['project_web_url'] + f'/-/issues/{issue_event.target_iid}',

                               # N.B.: project information is unable, and must be added by caller via kwargs.
                           } | kwargs)

    def _extract_commit_info(self, commit) -> LynceusDict:
        return LynceusDict({
            'id': commit.id,
            'short_id': commit.short_id,
            'parent_ids': commit.parent_ids,
            'message': commit.message,
            'created_at': commit.created_at,
            'author_name': commit.author_name,
            'author_email': commit.author_email,
            'committer_name': commit.committer_name,
            'committer_email': commit.committer_email,
        })

    def _extract_branch_info(self, branch) -> LynceusDict:
        return LynceusDict({
            'name': branch.name,
            'merged': branch.merged,
            'commit_id': branch.commit['id'],
            'commit_short_id': branch.commit['short_id'],
            'created_at': parse_string_to_datetime(datetime_str=branch.commit['created_at']),
            # Not available for other DevOps: 'project_id': branch.project_id,
        })

    def _extract_tag_info(self, tag) -> LynceusDict:
        return LynceusDict({
            'name': tag.name,
            'commit_id': tag.commit['id'],
            'commit_short_id': tag.commit['short_id'],
            'created_at': parse_string_to_datetime(datetime_str=tag.commit['created_at']),
            # Not available for other DevOps: 'project_id': branch.project_id,
        })

    # The following methods are only performing read access on DevOps backend.
    @gitlab_exception_handler
    def authenticate(self):
        self.__manager.auth()

    @gitlab_exception_handler
    def _do_get_current_user(self):
        return self.__manager.user

    @gitlab_exception_handler
    def _do_get_user_without_cache(self, *, username: str = None, email: str = None, **kwargs):
        users = self.__manager.users.list(iterator=True, username=username, email=email, **kwargs)
        try:
            user = next(users)
            self._logger.debug(f'Successfully lookup user with parameters: "{username=}"/"{email=}".')
            return user
        except StopIteration:
            # pylint: disable=raise-missing-from
            raise NameError(
                f'User "{username=}"/"{email=}" has not been found with specified parameters ({kwargs if kwargs else "none"}).')

    @gitlab_exception_handler
    def _do_get_groups(self, *, count: int | None = None, **kwargs):
        kwargs = {'all': True} | kwargs
        if count:
            return self.__manager.groups.list(**kwargs)[:count]
        return self.__manager.groups.list(**kwargs)

    @gitlab_exception_handler
    def _do_get_group_without_cache(self, *, full_path: str, **kwargs):
        groups = self.__manager.groups.list(iterator=True, search=full_path, search_namespaces=True, **kwargs)
        if not groups:
            raise NameError(
                f'Group "{full_path}" has not been found with specified parameters ({kwargs if kwargs else "none"}).')

        # Checks if match must be exact (which is now the case by default ... check if has been defined to False manually).
        if not bool(kwargs.get('exact_match', True)):
            if len(groups) > 1:
                self._logger.warning(f'There are more than one group matching requested path "{full_path}"' +
                                     f'(add "exact_match" parameter to ensure getting only one group): {groups} ')
            return next(groups)

        try:
            group = next(filter(lambda grp: str.lower(grp.full_path) == str.lower(full_path), groups))
            self._logger.debug(f'Successfully lookup group with full path "{full_path}".')
            return group
        except StopIteration:
            # pylint: disable=raise-missing-from
            raise NameError(
                f'Group "{full_path}" has not been found with specified parameters ({kwargs if kwargs else "none"}).' +
                f' Maybe you are looking for one of the following paths: {[group.full_path for group in groups]}')

    @gitlab_exception_handler
    def _do_get_projects(self, **kwargs):
        kwargs = {'all': True} | kwargs
        return self.__manager.projects.list(**kwargs)

    @gitlab_exception_handler
    def _do_get_project_without_cache(self, *, full_path: str, **kwargs):
        projects = self.__manager.projects.list(iterator=True, search=full_path, search_namespaces=True, **kwargs)
        try:
            project = next(projects)
            self._logger.debug(f'Successfully lookup project with full path "{full_path}".')
            return project
        except StopIteration:
            # pylint: disable=raise-missing-from
            raise NameError(
                f'Project "{full_path}" has not been found with specified parameters ({kwargs if kwargs else "none"}).')

    @gitlab_exception_handler
    def __get_current_token(self):
        if not self.__current_token:
            # Sample: {'id': 60, 'name': 'Lynceus CI/CD', 'revoked': False, 'created_at': '2020-10-22T15:50:13.516Z',
            #           'scopes': ['api', 'admin_mode'], 'user_id': 5, 'last_used_at': '2023-04-12T09:22:19.495Z',
            #           'active': True, 'expires_at': None}
            self.__current_token = self.__manager.personal_access_tokens.get('self')

        return self.__current_token

    @gitlab_exception_handler
    # pylint: disable=too-many-return-statements
    def check_permissions_on_project(self, *, full_path: str, get_metadata: bool, pull: bool,
                                     push: bool = False, maintain: bool = False, admin: bool = False, **kwargs):
        try:
            # First of all, checks get_metadata permission which is required anyway.
            _ = self._do_get_project(full_path=full_path, **kwargs)

            # From here, we consider get_metadata permission is OK.

            # Cf. https://docs.gitlab.com/ee/user/permissions.html
            # Cf. https://gitlab.com/gitlab-org/gitlab/-/blob/e97357824bedf007e75f8782259fe07435b64fbb/lib/gitlab/access.rb#L12-18

            # Retrieves current use, and corresponding membership information if any, for specified project.
            current_user = self._do_get_current_user()
            members = self._do_get_project_members(full_path=full_path, recursive=True)

            member = next(filter(lambda grp: grp.id == current_user.id, members))
            access_level = member.access_level

            # According to my tests, required role (aka access_level):
            #  - at least REPORTER role to get access to repository metadata (name, tags, branches ...)
            #  - at least DEVELOPER role to be able to pull the repository (the read_repository scope is NOT enough here)
            #  - at least MAINTAINER role to get access to project statistics

            # Checks scopes of the token.
            token = self.__get_current_token()
            token_scopes = token.scopes
            api_scope = 'api' in token_scopes
            admin_scope = 'admin' in token_scopes
            read_repository_scope = 'read_repository' in token_scopes
            write_repository_scope = 'write_repository' in token_scopes

            if pull and ((not api_scope and not read_repository_scope) or access_level < AccessLevel.DEVELOPER):
                return False

            if push and ((not api_scope and not write_repository_scope) or access_level < AccessLevel.DEVELOPER):
                return False

            if maintain and ((not api_scope and not write_repository_scope) or access_level < AccessLevel.MAINTAINER):
                return False

            if admin and ((not api_scope and not admin_scope) or access_level < AccessLevel.OWNER):
                return False

            # All permission checks OK.
            return True
        except NameError:
            # Returns True if there were NO permission at all to check ...
            return not get_metadata and not pull and not push and not maintain and not admin
        except StopIteration:
            # Returns True if there were NO more permission to check ...
            return not pull and not push and not maintain and not admin

    @gitlab_exception_handler
    def _do_get_project_commits(self, *, full_path: str, git_ref_name: str, count: int | None = None, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        if count:
            return project.commits.list(all=True, ref_name=git_ref_name)[:count]
        return project.commits.list(all=True, ref_name=git_ref_name)

    @gitlab_exception_handler
    def _do_get_project_branches(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        return project.branches.list(all=True)

    @gitlab_exception_handler
    def _do_get_project_tags(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        return project.tags.list(all=True)

    @gitlab_exception_handler
    def _do_get_project_members(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)

        if not bool(kwargs.get('recursive', False)):
            return project.members.list(all=True)

        return project.members_all.list(all=True)

    @gitlab_exception_handler
    def _do_get_group_members(self, *, full_path: str, **kwargs):
        group = self._do_get_group(full_path=full_path, **kwargs)

        if not bool(kwargs.get('recursive', False)):
            return group.members.list(all=True)

        return group.members_all.list(all=True)

    @gitlab_exception_handler
    def _do_get_project_issue_events(self, *, full_path: str, action: str | None = None,
                                     from_date: datetime | None = None,
                                     to_date: datetime | None = None, **kwargs):
        # Cf. https://python-gitlab.readthedocs.io/en/stable/api/gitlab.v4.html?highlight=events#gitlab.v4.objects.ProjectEventManager
        # Object listing filters
        #   action => https://docs.gitlab.com/ee/user/profile/index.html#user-contribution-events
        #   target_type => https://docs.gitlab.com/ee/api/events.html#target-types
        #   sort
        project = self._do_get_project(full_path=full_path, **kwargs)

        # Adds filters if needed.
        if action:
            kwargs['action'] = action

        #   before & after => https://docs.gitlab.com/ee/api/events.html#date-formatting
        if from_date:
            kwargs['after'] = from_date.strftime('%Y-%m-%d')

        if to_date:
            kwargs['before'] = to_date.strftime('%Y-%m-%d')

        return project.events.list(all=True, target_type='issue', **kwargs)

    @gitlab_exception_handler
    def _do_get_project_issues(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        return project.issues.list(all=True)

    @gitlab_exception_handler
    def _do_get_project_merge_requests(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        return project.mergerequests.list(all=True)

    @gitlab_exception_handler
    def _do_get_project_milestones(self, *, full_path: str, **kwargs):
        project = self._do_get_project(full_path=full_path, **kwargs)
        return project.milestones.list(all=True)

    @gitlab_exception_handler
    def __get_recursive_groups(self, *, full_path: str, **kwargs):
        group = self._do_get_group(full_path=full_path, **kwargs)
        subgroups = group.subgroups.list(all=True)
        if not subgroups:
            return [group]
        return flatten(self.__get_recursive_groups(full_path=subgroup.full_path, **kwargs) for subgroup in subgroups)

    @gitlab_exception_handler
    def _do_get_group_projects(self, *, full_path: str, **kwargs):
        group = self._do_get_group(full_path=full_path, **kwargs)

        # Checks if recursive is requested.
        if not bool(kwargs.get('recursive', False)):
            project_list = group.projects.list(all=True)
        else:
            all_groups = {group} | set(self.__get_recursive_groups(full_path=full_path, **kwargs))
            project_list = flatten(group.projects.list(all=True) for group in all_groups)

        return list(map(lambda gp: self._do_get_project(full_path=gp.path_with_namespace), project_list))

    @gitlab_exception_handler
    def _do_get_group_milestones(self, *, full_path: str, **kwargs):
        group = self._do_get_group(full_path=full_path, **kwargs)
        return group.milestones.list(all=True)

    @gitlab_exception_handler
    def get_user_stats_commit_activity(self, *, group_full_path: str = None, project_full_path: str = None,
                                       since: datetime = None, keep_empty_stats: bool = False):
        # Cf. https://docs.gitlab.com/ee/api/project_statistics.html#get-the-statistics-of-the-last-30-days

        # Defines projects on which to perform statistics.
        if group_full_path is None:
            projects = self._do_get_projects()
        else:
            projects = self._do_get_group_projects(full_path=group_full_path)
            if project_full_path is not None:
                projects.append(self._do_get_project(full_path=project_full_path))

        # Defines threshold date.
        contributions_since: datetime = since if since else datetime.now(tz=timezone.utc) - timedelta(days=365)
        stats_user_commit_activity: dict[datetime.date, int] = defaultdict(int)
        for _project in projects:
            for commit_activity in _project.additionalstatistics.get().fetches['days']:
                day_date = parse_string_to_datetime(datetime_str=commit_activity['date'], datetime_format=DATETIME_FORMAT_SHORT)

                # Ignores oldest statistics.
                if day_date < contributions_since:
                    continue

                # Ignores 0 stats but if wanted.
                if not keep_empty_stats and not commit_activity['count']:
                    continue

                stats_user_commit_activity[day_date.date()] += commit_activity['count']

        return stats_user_commit_activity

    @gitlab_exception_handler
    # pylint: disable=unused-argument
    def get_user_contributions(self, *, since: datetime = None, keep_empty_stats: bool = False):
        self._logger.warning('get_user_contributions is not implemented yet for GitLab DevOps Analyzer')
        return {}

    @gitlab_exception_handler
    def get_user_stats_code_frequency(self):
        self._logger.warning(
            'get_user_stats_commit_activity is not implemented yet for GitLab DevOps Analyzer')
        return {}

    @gitlab_exception_handler
    def _do_download_repository(self, *, project_full_path: str, dest_path: Path,
                                reference: str = None, chunk_size: int = 1024, **kwargs):
        # Cf. https://docs.gitlab.com/ee/api/repositories.html#get-file-archive
        # Cf. https://github.com/python-gitlab/python-gitlab/blob/main/gitlab/v4/objects.py#L4465
        project = self._do_get_project(full_path=project_full_path, **kwargs)
        with open(dest_path, 'wb') as dest_file:
            project.repository_archive(sha=reference, streamed=True, action=dest_file.write, chunk_size=chunk_size,
                                       **kwargs)

    # The following methods are performing write/delete operation on DevOps backend.
    @gitlab_exception_handler
    def _do_create_group(self, *, parent_group_full_path: str, new_group_name: str, new_group_relative_path: str, **kwargs):
        # Prepares the common parameters.
        group_create_params: dict[str, str] = {'name': new_group_name,
                                               'path': new_group_relative_path,
                                               **kwargs}

        # Checks if the group must be created under an existing one.
        if parent_group_full_path:
            root_group = self._do_get_group(full_path=parent_group_full_path, exact_match=True)
            if not root_group:
                raise NameError(f'Unable to find the parent group with name "{parent_group_full_path}".')

            # Adds the parent parameter.
            group_create_params.update({'parent_id': root_group.id})

        # Cf. https://docs.gitlab.com/ee/api/groups.html#new-group
        # Requests group creation.
        return self.__manager.groups.create(group_create_params)

    @gitlab_exception_handler
    def _do_create_user(self, *, name: str, username: str, email: str, **kwargs):
        # Cf. https://python-gitlab.readthedocs.io/en/stable/gl_objects/users.html
        return self.__manager.users.create(
            {
                'name': name,
                'username': username,
                'email': email,
                'reset_password': True,
                **kwargs})

    @gitlab_exception_handler
    def _do_update_user_notification_settings(self, *, username: str, email: str, **kwargs):
        # Cf. https://python-gitlab.readthedocs.io/en/stable/gl_objects/notifications.html

        # N.B.: in Gitlab, only the notification settings of the authenticated user can be updated.
        # so here, to update the settings of the specified user, we need to request a dedicated impersonation token.
        # Cf. https://python-gitlab.readthedocs.io/en/stable/gl_objects/users.html?highlight=impersonation#user-impersonation-tokens
        user = self._do_get_user(username=username, email=email)

        # Creates a new impersonation token.
        user_active_impersonation_tokens = user.impersonationtokens.list(state='active')
        self._logger.debug(f'Looking for existing "active" impersonation token for user "{self._extract_user_info(user)}": "{user_active_impersonation_tokens=}".')
        impersonation_token = user.impersonationtokens.create({'name': 'notif_token', 'scopes': ['api']})
        self._logger.debug(f'Successfully created new impersonation token for user "{self._extract_user_info(user)}": "{impersonation_token=}".')

        # Creates a new dedicated Gitlab manager, to update the settings.
        user_gl = gitlab.Gitlab(self._uri, private_token=impersonation_token.token)
        user_notificationsettings = user_gl.notificationsettings.get()
        self._logger.debug(f'These are the current user notification settings: "{user_notificationsettings.attributes=}".')

        # Updates notification settings.
        for setting, value in kwargs.items():
            self._logger.debug(f'Updating user notification settings "{setting}" to "{value}".')
            setattr(user_notificationsettings, setting, value)
        user_notificationsettings.save()
        self._logger.debug(f'Successfully updated user notification settings to: "{user_notificationsettings.attributes=}".')

        # Cleans the impersonation token.
        impersonation_token.delete()
        del user_gl

        # Returns the User which may hold some interesting information in some DevOps Backend implementation.
        return user

    @gitlab_exception_handler
    def _do_add_group_member(self, *, group_full_path: str, username: str, access, **kwargs):
        group = self._do_get_group(full_path=group_full_path, exact_match=True)
        user = self._do_get_user(username=username)

        return group.members.create(
            {'user_id': user.id,
             'access_level': access,
             **kwargs})

    def __wait_until_status_or_timeout(self, *, obj, refresh_func, attr_name: str, values: list[str], timeout_sec: int, step_sec: int = 2):
        self._logger.debug(f'Starting polling wait until maximum of {timeout_sec} seconds, for attributes "{attr_name=}" on "{obj=}" to reach one of the "{values=}" ...')
        remaining_time: int = timeout_sec
        refresh_func(obj)
        while remaining_time > 0 and getattr(obj, attr_name) not in values:
            time.sleep(step_sec)
            remaining_time -= step_sec
            refresh_func(obj)

        self._logger.debug(f'Attribute "{attr_name=}" on "{obj=}" reached "{getattr(obj, attr_name)}", in "{timeout_sec - remaining_time}" seconds.')

    @gitlab_exception_handler
    def _do_export_project(self, *, project_full_path: str, export_dst_full_path: Path, timeout_sec: int = 60, **kwargs):
        # Cf. https://python-gitlab.readthedocs.io/en/stable/gl_objects/projects.html#import-export
        # Cf. https://docs.gitlab.com/ce/api/project_import_export.html

        # Safe-guard: prepares output structure.
        export_dst_full_path.parent.mkdir(parents=True, exist_ok=True)

        # Create the export.
        self._logger.info(f'Requesting export of project with full path "{project_full_path}" ...')
        project = self._do_get_project(full_path=project_full_path)
        export = project.exports.create()

        # Waits till the export reaches a final status.
        self.__wait_until_status_or_timeout(obj=export,
                                            refresh_func=lambda instance: instance.refresh(),
                                            attr_name='export_status',
                                            values=[self.IMPORT_EXPORT_STATUS_SUCCESS, self.IMPORT_EXPORT_STATUS_FAILED],
                                            timeout_sec=timeout_sec)

        # Checks status.
        if export.export_status != self.IMPORT_EXPORT_STATUS_SUCCESS:
            raise ValueError(f'Export of project with full path "{project_full_path}" did not reach' +
                             f' "{self.IMPORT_EXPORT_STATUS_SUCCESS}" success status, in "{timeout_sec}" seconds' +
                             f' but "{export.export_status}".')

        # Saves the result.
        self._logger.debug(f'Export is now ready, saving it to "{export_dst_full_path}" ...')
        with open(export_dst_full_path, 'wb') as export_file:
            export.download(streamed=True, action=export_file.write)
        self._logger.info(f'Successfully exported project with full path "{project_full_path}" to "{export_dst_full_path}".')

    @gitlab_exception_handler
    def _do_import_project(self, *, parent_group_full_path: str,
                           new_project_name: str, new_project_path: str,
                           import_src_full_path: Path, timeout_sec: int = 60, **kwargs):
        # Cf. https://python-gitlab.readthedocs.io/en/stable/gl_objects/projects.html#import-export
        # Cf. https://docs.gitlab.com/ce/api/project_import_export.html

        # Safe-guard: ensures import source file exists..
        if not import_src_full_path.exists():
            raise ValueError(f'Specified source file "{import_src_full_path}" does not exist. Aborting import.')

        # Create the import.
        project_complete_path = new_project_path if not parent_group_full_path else f'{parent_group_full_path}/{new_project_path}'
        self._logger.info(f'Requesting import of project "{import_src_full_path}" to new project with full path "{project_complete_path}" ...')

        with open(import_src_full_path, 'rb') as file_source:
            output = self.__manager.projects.import_project(
                file_source,
                namespace=parent_group_full_path,
                path=new_project_path, name=new_project_name,
                **kwargs)
        new_project = self.__manager.projects.get(output['id'])
        project_import = new_project.imports.get()

        # Waits till the import reaches a final status.
        self.__wait_until_status_or_timeout(obj=project_import,
                                            refresh_func=lambda instance: instance.refresh(),
                                            attr_name='import_status',
                                            values=[self.IMPORT_EXPORT_STATUS_SUCCESS, self.IMPORT_EXPORT_STATUS_FAILED],
                                            timeout_sec=timeout_sec)

        # Checks status.
        if project_import.import_status != self.IMPORT_EXPORT_STATUS_SUCCESS:
            raise ValueError(f'Import of "{import_src_full_path}" to new project with full path "{new_project.path_with_namespace}" did not reach' +
                             f' "{self.IMPORT_EXPORT_STATUS_SUCCESS}" success status, in "{timeout_sec}" seconds' +
                             f' but "{project_import.import_status}".')

        self._logger.info(f'Successfully imported "{import_src_full_path}" to new project with full path "{new_project.path_with_namespace}" ...')
